from django.contrib.auth.base_user import BaseUserManager
from django.contrib.auth.models import AbstractUser
from django.contrib.auth.validators import UnicodeUsernameValidator
from django.core.validators import EmailValidator
from django.db import models
from django.utils.translation import gettext_lazy as _


class UserManager(BaseUserManager):
    """Define a model manager for User model with no username field."""
    use_in_migrations = True

    def _create_user(self, email, password, **extra_fields):
        """Create and save a User with the given email and password."""
        if not email:
            raise ValueError(_('The given email must be set'))
        email = self.normalize_email(email)
        user = self.model(email=email, **extra_fields)
        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_user(self, email, password=None, **extra_fields):
        """Create and save a regular User with the given email and password."""
        extra_fields.setdefault('is_staff', False)
        extra_fields.setdefault('is_superuser', False)
        return self._create_user(email, password, **extra_fields)

    def create_superuser(self, email, password, **extra_fields):
        """Create and save a SuperUser with the given email and password."""
        extra_fields.setdefault('is_staff', True)
        extra_fields.setdefault('is_superuser', True)

        if extra_fields.get('is_staff') is not True:
            raise ValueError('Superuser must have is_staff=True.')
        if extra_fields.get('is_superuser') is not True:
            raise ValueError('Superuser must have is_superuser=True.')

        return self._create_user(email, password, **extra_fields)


class User(AbstractUser):
    objects = UserManager()

    USERNAME_FIELD = 'username'
    EMAIL_FIELD = 'email'
    REQUIRED_FIELDS = ['phone']
    AUTHENTICATION_FIELDS = ['email', 'username', 'phone', 'id']

    username_validator = UnicodeUsernameValidator()
    email_validator = EmailValidator()

    email = models.EmailField(_('email address'),
                              validators=[email_validator],
                              unique=True, blank=False, null=False)
    email_verified_at = models.DateField(blank=True, null=True)

    username = models.CharField(_('username'), max_length=150,
                                help_text=_('Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.'),
                                validators=[username_validator], unique=True)
    phone = models.CharField(
        max_length=50,
        blank=False,
        null=False,
        unique=True,
        error_messages={'unique': _("A user with that phone already exists.")})

    phone_verified_at = models.DateTimeField(blank=True, null=True)

    def __str__(self):
        return self.username

    def clean(self):
        self.email = self.email.lower()



from django.urls import path, include

urlpatterns = [
    # add account profile
    path('', include('dj_accounts.urls.site_urls.urls_profile')),
    path('', include('dj_accounts.urls.site_urls.urls_auth')),
    # include api urls
    path('api/', include('dj_accounts.urls.api_urls.urls_auth_api')),
    path('api/', include('dj_accounts.urls.api_urls.urls_profile_api')),
]



from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth import password_validation
from django.contrib.auth.forms import PasswordResetForm, AuthenticationForm, PasswordChangeForm
from django.contrib.auth.password_validation import validate_password
from django.utils.translation import gettext as _
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from rest_framework_simplejwt.exceptions import TokenError
from rest_framework_simplejwt.tokens import RefreshToken

from .forms import RegisterForm, MultipleLoginForm

UserModel = get_user_model()


class LoginSerializer(serializers.Serializer):
    email = serializers.EmailField(required=False, allow_blank=True)
    username = serializers.CharField(required=False, allow_blank=True)
    phone = serializers.CharField(required=False, allow_blank=True)
    password = serializers.CharField(write_only=True)
    cache_user = None

    def validate(self, attrs):
        form = self.get_authentication_form()(data=dict(**attrs))
        if not form.is_valid():
            raise serializers.ValidationError(form.errors)

        self.cache_user = form.get_user()
        return attrs

    def get_authentication_form(self):
        if getattr(settings, 'MULTIPLE_AUTHENTICATION_ACTIVE', False):
            return MultipleLoginForm
        return getattr(settings, 'LOGIN_FORM', AuthenticationForm)


class LogoutSerializer(serializers.Serializer):
    refresh = serializers.CharField(required=True, help_text=_("Required, please provide your refresh token"))

    def save(self, **kwargs):
        try:
            RefreshToken(self.validated_data['refresh']).blacklist()
        except TokenError:
            raise ValidationError({"refresh": _('Invalid token.')})


class RegisterSerializer(serializers.ModelSerializer):
    username = serializers.CharField(
        required=True,
        max_length=100, help_text=_("Required, please provide your username"))

    email = serializers.EmailField(
        required=True,
        max_length=100, help_text=_("Required, please provide your email"))

    phone = serializers.CharField(
        required=True,
        max_length=100, help_text=_("Required, please provide your phone number"))

    password1 = serializers.CharField(
        required=True,

        help_text=password_validation.password_validators_help_text_html())

    password2 = serializers.CharField(
        required=True,
        help_text=_("Enter the same password as before, for verification."))

    class Meta:
        model = UserModel
        fields = ['username', 'email', 'password1', 'password2', 'phone']

    def validate(self, attrs):
        form = RegisterForm(data=attrs)
        if not form.is_valid():
            raise serializers.ValidationError(form.errors)
        return attrs

    def create(self, validated_data):
        validated_data.pop('password2')
        password = validated_data.pop('password1')
        user = super(RegisterSerializer, self).create(validated_data)
        user.set_password(password)
        user.save()
        return user


class PasswordResetSerializer(serializers.Serializer):
    email = serializers.EmailField(max_length=254)

    form = None

    def validate(self, attrs):
        self.form = PasswordResetForm(data=attrs)
        if not self.form.is_valid():
            raise serializers.ValidationError(self.form.errors)
        return attrs

    def save(self, **kwargs):
        opts = kwargs.get('opts')
        self.form.save(**opts)


class UpdateUserDataSerializer(serializers.ModelSerializer):
    class Meta:
        model = get_user_model()
        fields = ['first_name', 'last_name']


class UpdateEmailSerializer(serializers.ModelSerializer):
    class Meta:
        model = get_user_model()
        fields = ['email', 'password']


class UpdatePhoneNumberSerializer(serializers.ModelSerializer):
    class Meta:
        model = get_user_model()
        fields = ['phone', 'password']


class ChangePasswordSerializer(serializers.ModelSerializer):
    class Meta:
        model = UserModel
        fields = ['new_password1', 'new_password2', 'old_password']

    new_password1 = serializers.CharField(required=True, write_only=True)
    new_password2 = serializers.CharField(required=True, write_only=True, validators=[validate_password])
    old_password = serializers.CharField(required=True, write_only=True)
    form = None

    def validate(self, data):
        self.form = PasswordChangeForm(data=data, user=self.context['request'].user)
        if not self.form.is_valid():
            serializers.ValidationError(self.form.errors)
        return data

    def save(self, **kwargs):
        self.form.save()



import inspect

from django.core import mail
from django.test import TestCase, override_settings
from django.urls import reverse
from django.utils import timezone, dateformat
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from rest_framework import status
from rest_framework.permissions import IsAuthenticated
from rest_framework.reverse import reverse
from rest_framework.test import APIClient
from rest_framework.test import APITestCase
from rest_framework.views import APIView
from rest_framework_simplejwt.tokens import RefreshToken

from .factories import UserFactory
from ..serializers import ChangePasswordSerializer
from ..utils import account_activation_token
from ..views_api import UpdateProfileAPIView, ChangeEmailAPIView, ChangePhoneAPIView, \
    VerifyPhoneAPIView, VerifyEmailAPIView, ResendPhoneConfirmationAPIView, UserLogoutAPIView, ChangePasswordAPIView


# test auth
class LoginAPIViewTestCase(TestCase):
    def setUp(self):
        self.client = APIClient()
        self.user = UserFactory(username="TestTest", email="test@aol.com", phone="+123456789")
        self.url = reverse("login_api")
        self.data = {
            "username": "TestTest",
            "email": "test@aol.com",
            "phone": "+123456789",
            "password": "secret"
        }

    @override_settings(MULTIPLE_AUTHENTICATION_ACTIVE=True)
    def test_it_uses_email_and_password_for_successful_login(self):
        self.data.pop("phone")
        self.data.pop("username")
        response = self.client.post(self.url, self.data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @override_settings(MULTIPLE_AUTHENTICATION_ACTIVE=True)
    def test_it_uses_username_and_password_for_successful_login(self):
        self.data.pop("phone")
        self.data.pop("email")
        response = self.client.post(self.url, self.data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @override_settings(MULTIPLE_AUTHENTICATION_ACTIVE=True,
                       AUTHENTICATION_BACKENDS=["dj_accounts.backends.MultipleAuthenticationBackend"])
    def test_it_uses_phone_and_password_for_successful_login(self):
        self.data.pop("email")
        self.data.pop("username")
        response = self.client.post(self.url, self.data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @override_settings(MULTIPLE_AUTHENTICATION_ACTIVE=True)
    def test_it_returns_access_token_key_in_response(self):
        self.data.pop("phone")
        self.data.pop("username")
        response = self.client.post(self.url, self.data)
        self.assertIn("access_token", response.data)

    @override_settings(MULTIPLE_AUTHENTICATION_ACTIVE=True)
    def test_it_returns_refresh_token_key_in_response(self):
        self.data.pop("phone")
        self.data.pop("username")
        response = self.client.post(self.url, self.data)
        self.assertIn("refresh_token", response.data)

    def test_it_returns_401_unauthenticated_if_credentials_are_invalid(self):
        self.data["password"] = "invalid"
        response = self.client.post(self.url, self.data)
        self.assertEqual(response.status_code, status.HTTP_422_UNPROCESSABLE_ENTITY)

    def test_it_returns_validation_errors_if_credentials_are_invalid(self):
        self.data["password"] = "invalid"
        response = self.client.post(self.url, self.data)
        self.assertIn("__all__", response.data)

    def test_it_returns_password_in_validation_errors(self):
        self.data.pop("password")
        response = self.client.post(self.url, self.data)
        self.assertIn("password", response.data)


# class LoginAPIViewTestCase(TestCase):
#     def setUp(self):
#         self.url = reverse('api-v1:login')
#         self.user = UserFactory()
#         self.client = APIClient()
#
#     def test_it_return_401_if_not_active_user_tried_to_authenticate(self):
#         self.user.is_active = False
#         self.user.save()
#         response = self.client.post(self.url, {'email': self.user.email, 'password': '123'})
#         self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
#
#     def test_it_return_401_if_user_does_not_exist(self):
#         response = self.client.post(self.url, {'email': 'not_exist', 'password': '123'})
#         self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
#
#     def test_it_return_401_if_password_is_wrong(self):
#         response = self.client.post(self.url, {'email': self.user.email, 'password': 'wrong'})
#         self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
#
#     def test_it_return_200_if_user_is_valid(self):
#         response = self.client.post(self.url, {'email': self.user.email, 'password': 'secret'})
#         self.assertEqual(response.status_code, status.HTTP_200_OK)
#
#     def test_it_return_access_token_after_user_is_authenticated_correctly(self):
#         response = self.client.post(self.url, {'email': self.user.email, 'password': 'secret'})
#         self.assertTrue('access' in response.data)
#
#     def test_it_return_refresh_token_after_user_is_authenticated_correctly(self):
#         response = self.client.post(self.url, {'email': self.user.email, 'password': 'secret'})
#         self.assertTrue('refresh' in response.data)
#
#     def test_it_return_401_if_invalid_token_was_given(self):
#         client = APIClient()
#         client.credentials(HTTP_AUTHORIZATION='Bearer ' + 'abc')
#         response = client.get(reverse('api-v1:verify-phone'))
#         self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
#
#     def test_the_returned_token_is_valid(self):
#         client = APIClient()
#         response = client.post(self.url, {'email': self.user.email, 'password': 'secret'})
#         self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
#         token = response.data['access']
#         client.credentials(HTTP_AUTHORIZATION='Bearer {}'.format(token))
#         response = client.get(reverse('api-v1:resend-email-activation'))
#         self.assertEqual(response.status_code, status.HTTP_200_OK)


# logout

class TestUserLogoutViewStructure(TestCase):
    def test_it_extends_api_view_class(self):
        self.assertTrue(issubclass(UserLogoutAPIView, APIView))

    def test_view_has_method_post(self):
        self.assertTrue(hasattr(UserLogoutAPIView, 'post'))

    def test_view_has_method_post_is_callable(self):
        self.assertTrue(callable(UserLogoutAPIView.post))

    def test_post_method_signature(self):
        expected_signature = ['self', 'request']
        actual_signature = inspect.getfullargspec(UserLogoutAPIView.post)[0]
        self.assertEquals(actual_signature, expected_signature)


class TestUserLogoutView(TestCase):
    @override_settings(MULTIPLE_AUTHENTICATION_ACTIVE=True)
    def setUp(self):
        self.url = reverse('login_api')
        self.user = UserFactory()
        self.client = APIClient()
        login_response = self.client.post(self.url, {'email': self.user.email, 'password': 'secret'})
        self.refresh = login_response.data['refresh_token']
        self.token = login_response.data['access_token']

    def test_it_return_401_if_user_not_logged_in(self):
        response = self.client.post(reverse('logout_api'), {'refresh': self.refresh})
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_it_return_204_if_user_is_logged_out(self):
        self.client.credentials(HTTP_AUTHORIZATION='Bearer {}'.format(self.token))
        response = self.client.post(reverse('logout_api'), {'refresh': self.refresh})
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_it_return_400_if_invalid_token_was_given(self):
        self.client.credentials(HTTP_AUTHORIZATION='Bearer {}'.format(self.token))
        response = self.client.post(reverse('logout_api'), {'refresh': 'abc'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)


class UserSignupAPIViewTestCase(TestCase):
    def setUp(self):
        self.client = APIClient()
        self.user = UserFactory()
        self.url = reverse('signup_api')
        self.data = {
            "email": "test@test.test",
            "username": "TestUser",
            "phone": "+201005263988",
            "password1": "newTESTPasswordD",
            "password2": "newTESTPasswordD",
            "gender": "male",
            "birthdate": dateformat.format(timezone.now(), 'Y-m-d'),
        }

    def test_it_returns_422_when_data_is_invalid(self):
        """
            set Up :
              - we are trying post empty registration form data to serializer
            result : returning response 422 the data you posted not valid
        """
        response = self.client.post(self.url, {})
        self.assertEquals(response.status_code, 422)

    def test_it_returns_201_when_user_created_successfully(self):
        """
              set Up :
                - we are trying post registration data to serializer
              result : returning response 201 the data you posted is right and create product
        """
        response = self.client.post(self.url, self.data)
        self.assertEquals(response.status_code, 201)

    def test_it_create_send_the_email(self):
        response = self.client.post(self.url, self.data)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].subject, 'Activate your account.')

    def test_it_return_access_and_refresh_tokens_once_user_is_signup(self):
        response = self.client.post(self.url, self.data)
        self.assertIn('access_token', response.data)
        self.assertIn('refresh_token', response.data)


# phone verification
class VerifyPhoneAPIViewStructureTestCase(TestCase):
    def test_it_extends_django_view_class(self):
        self.assertTrue(issubclass(VerifyPhoneAPIView, APIView))

    def test_view_has_method_post(self):
        self.assertTrue(hasattr(VerifyPhoneAPIView, 'post'))

    def test_view_has_method_post_is_callable(self):
        self.assertTrue(callable(VerifyPhoneAPIView.post))

    def test_post_method_signature(self):
        expected_signature = ['self', 'request']
        actual_signature = inspect.getfullargspec(VerifyPhoneAPIView.post)[0]
        self.assertEquals(actual_signature, expected_signature)


class VerifyPhoneAPIViewPOSTTestCase(TestCase):
    def setUp(self):
        self.client = APIClient()
        self.user = UserFactory(phone="12312123")
        self.refresh = RefreshToken.for_user(self.user)
        self.client.credentials(HTTP_AUTHORIZATION='Bearer {}'.format(self.refresh.access_token))
        self.url = reverse("verify_phone_api")
        self.data = {"code": "777777"}

    def test_it_return_401_status_code_if_user_is_not_logged_in(self):
        self.client.logout()
        response = self.client.post(self.url)
        self.assertEquals(response.status_code, 401)

    def test_it_return_200_if_user_is_verified(self):
        user = UserFactory(phone_verified_at=now())
        self.client.login(email=user.email, password="secret")
        response = self.client.post(self.url, self.data)
        self.assertEquals(response.status_code, 200)

    def test_it_updates_phone_verified_at_column_in_user_model_to_now_on_success(self):
        self.client.post(self.url, self.data)
        self.user.refresh_from_db()
        self.assertIsNotNone(self.user.phone_verified_at)

    def test_it_raise_422_status_code_for_invalid_data(self):
        self.data = {"code": ""}
        response = self.client.post(self.url, self.data)
        self.assertEquals(response.status_code, 422)

    def test_it_return_400_bad_request_user_was_verified_before(self):
        self.user.phone_verified_at = now()
        self.user.save()
        response = self.client.post(self.url, self.data)
        self.assertEquals(response.status_code, 400)


class ResendPhoneConfirmationViewStructureTestCase(TestCase):
    def test_it_extends_django_LoginView(self):
        self.assertTrue(issubclass(ResendPhoneConfirmationAPIView, APIView))

    def test_it_permission_classes_has_is_authenticated(self):
        self.assertIn(IsAuthenticated, ResendPhoneConfirmationAPIView.permission_classes)

    def test_view_has_method_get(self):
        self.assertTrue(hasattr(ResendPhoneConfirmationAPIView, 'get'))

    def test_view_has_method_get_is_callable(self):
        self.assertTrue(callable(ResendPhoneConfirmationAPIView.get))

    def test_get_method_signature(self):
        expected_signature = ['self', 'request']
        actual_signature = inspect.getfullargspec(ResendPhoneConfirmationAPIView.get)[0]
        self.assertEquals(actual_signature, expected_signature)


class PhoneConfirmationViewGETTestCase(TestCase):
    def setUp(self):
        self.client = APIClient()
        self.user = UserFactory()
        self.refresh = RefreshToken.for_user(self.user)
        self.client.credentials(HTTP_AUTHORIZATION='Bearer {}'.format(self.refresh.access_token))
        self.url = reverse("resend_phone_activation_api")

    def test_it_returns_status_code_of_401_if_user_is_not_authenticated(self):
        self.client.logout()
        response = self.client.get(self.url)
        self.assertEquals(response.status_code, 401)

    def test_it_return_200_status_code_if_code_was_resent_successfully(self):
        response = self.client.get(self.url)
        self.assertEquals(response.status_code, 200)

    def test_response_message_value(self):
        response = self.client.get(self.url)
        self.assertEqual(response.data['message'], _('Code was resent successfully.'))


# email verification

class VerifyEmailViewStructureTestCase(TestCase):
    def test_it_extends_django_view_class(self):
        self.assertTrue(issubclass(VerifyEmailAPIView, APIView))

    def test_view_has_method_get(self):
        self.assertTrue(hasattr(VerifyEmailAPIView, 'get'))

    def test_view_has_method_get_is_callable(self):
        self.assertTrue(callable(VerifyEmailAPIView.get))

    def test_get_method_signature(self):
        expected_signature = ['self', 'request', 'uidb64', 'token']
        actual_signature = inspect.getfullargspec(VerifyEmailAPIView.get)[0]
        self.assertEquals(actual_signature, expected_signature)


class VerifyEmailAPIViewTestCase(TestCase):
    def setUp(self):
        self.client = APIClient()
        self.user = UserFactory()
        self.refresh = RefreshToken.for_user(self.user)
        self.client.credentials(HTTP_AUTHORIZATION='Bearer {}'.format(self.refresh.access_token))
        self.uid = urlsafe_base64_encode(force_bytes(self.user.pk))
        self.token = account_activation_token.make_token(self.user)

    def test_it_return_200_status_code_when_email_was_confirmed_successfully(self):
        response = self.client.get(reverse('verify_email_api', args=[self.uid, self.token]))
        self.assertEquals(response.status_code, 200)

    def test_message_value_when_email_was_verified(self):
        response = self.client.get(reverse('verify_email_api', args=[self.uid, self.token]))
        self.assertEquals(response.data['message'], _('Email was verified successfully.'))


class UpdateProfileDataAPIViewStructureTestCase(TestCase):
    def test_it_extends_django_LoginView(self):
        self.assertTrue(issubclass(UpdateProfileAPIView, APIView))

    def test_it_permission_classes_has_is_authenticated(self):
        self.assertIn(IsAuthenticated, UpdateProfileAPIView.permission_classes)

    def test_view_has_method_put(self):
        self.assertTrue(hasattr(UpdateProfileAPIView, 'put'))

    def test_view_has_method_put_is_callable(self):
        self.assertTrue(callable(UpdateProfileAPIView.put))

    def test_put_method_signature(self):
        expected_signature = ['self', 'request']
        actual_signature = inspect.getfullargspec(UpdateProfileAPIView.put)[0]
        self.assertEquals(actual_signature, expected_signature)

    def test_it_has_get_serializer_class_method(self):
        self.assertIn('get_serializer_class', UpdateProfileAPIView.__dict__)


class UpdateProfileDataAPIViewTestCase(APITestCase):
    def setUp(self):
        self.client = APIClient()
        self.user = UserFactory()
        self.refresh = RefreshToken.for_user(self.user)
        self.client.credentials(HTTP_AUTHORIZATION='Bearer {}'.format(self.refresh.access_token))
        self.url = reverse('update-profile-api')
        self.data = {
            'first_name': self.user.first_name,
            'last_name': self.user.last_name,
        }

    def tearDown(self):
        self.client.logout()

    def test_it_through_exception_if_user_is_not_authenticated(self):
        self.client.logout()
        response = self.client.put(self.url)
        self.assertEquals(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_it_returns_status_code_of_201_if_data_profile_is_updated(self):
        response = self.client.put(self.url, self.data)
        self.assertEquals(response.status_code, status.HTTP_201_CREATED)

    def test_it_returns_user_data_after_updated(self):
        response = self.client.put(self.url, self.data)
        self.assertIsNotNone(response.data)

    def test_it_returns_201_when_user_created_successfully(self):
        response = self.client.put(self.url, data=self.data)
        self.assertEquals(response.status_code, 201)


class UpdateEmailAPIViewStructureTestCase(TestCase):
    def test_it_extends_django_LoginView(self):
        self.assertTrue(issubclass(ChangeEmailAPIView, APIView))

    def test_it_permission_classes_has_is_authenticated(self):
        self.assertIn(IsAuthenticated, ChangeEmailAPIView.permission_classes)

    def test_view_has_method_put(self):
        self.assertTrue(hasattr(ChangeEmailAPIView, 'put'))

    def test_view_has_method_put_is_callable(self):
        self.assertTrue(callable(ChangeEmailAPIView.put))

    def test_put_method_signature(self):
        expected_signature = ['self', 'request']
        actual_signature = inspect.getfullargspec(ChangeEmailAPIView.put)[0]
        self.assertEquals(actual_signature, expected_signature)

    def test_it_has_get_serializer_class_method(self):
        self.assertIn('get_serializer_class', ChangeEmailAPIView.__dict__)


class UpdateEmailAPIViewTestCase(APITestCase):
    def setUp(self):
        self.client = APIClient()
        self.user = UserFactory(email='test@test.com')
        self.refresh = RefreshToken.for_user(self.user)
        self.client.credentials(HTTP_AUTHORIZATION='Bearer {}'.format(self.refresh.access_token))
        self.url = reverse('change-email-api')
        self.data = {
            'new_email': "newtest@test.com",
            "password": "secret",
        }

    def tearDown(self):
        self.client.logout()

    def test_it_through_exception_if_user_is_not_authenticated(self):
        self.client.logout()
        response = self.client.put(self.url)
        self.assertEquals(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_it_returns_422_when_data_is_invalid(self):
        response = self.client.put(self.url, data={})
        self.assertEquals(response.status_code, 422)

    def test_it_returns_201_when_user_created_successfully(self):
        response = self.client.put(self.url, data=self.data)
        self.assertEquals(response.status_code, 201)

    def test_it_returns_422_if_user_tried_to_enter_the_same_email(self):
        data = {
            'email': "test@test.com",
            "password": "secret",
        }
        response = self.client.put(self.url, data=data)
        self.assertEquals(response.status_code, 422)

    def test_it_change_email_for_user(self):
        self.client.put(self.url, data=self.data)
        self.user.refresh_from_db()
        self.assertEquals(self.user.email, self.data['new_email'])


class UpdatePhoneAPIViewStructure(TestCase):
    def test_it_extends_django_LoginView(self):
        self.assertTrue(issubclass(ChangePhoneAPIView, APIView))

    def test_it_permission_classes_has_is_authenticated(self):
        self.assertIn(IsAuthenticated, ChangePhoneAPIView.permission_classes)

    def test_view_has_method_put(self):
        self.assertTrue(hasattr(ChangePhoneAPIView, 'put'))

    def test_view_has_method_put_is_callable(self):
        self.assertTrue(callable(ChangePhoneAPIView.put))

    def test_put_method_signature(self):
        expected_signature = ['self', 'request']
        actual_signature = inspect.getfullargspec(ChangePhoneAPIView.put)[0]
        self.assertEquals(actual_signature, expected_signature)

    def test_it_has_get_serializer_class_method(self):
        self.assertIn('get_serializer_class', ChangePhoneAPIView.__dict__)


class UpdatePhoneAPIViewTestCase(APITestCase):
    def setUp(self):
        self.client = APIClient()
        self.user = UserFactory(phone='+201005263977')
        self.refresh = RefreshToken.for_user(self.user)
        self.client.credentials(HTTP_AUTHORIZATION='Bearer {}'.format(self.refresh.access_token))
        self.url = reverse('change-phone-api')
        self.data = {
            'new_phone': "+201005263988",
            "password": "secret",
        }

    def tearDown(self):
        self.client.logout()

    def test_it_through_exception_if_user_is_not_authenticated(self):
        self.client.logout()
        response = self.client.put(self.url)
        self.assertEquals(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_it_returns_422_when_data_is_invalid(self):
        response = self.client.put(self.url, data={})
        self.assertEquals(response.status_code, 422)

    def test_it_returns_201_when_user_created_successfully(self):
        response = self.client.put(self.url, data=self.data)
        self.assertEquals(response.status_code, 201)

    def test_it_returns_201_when_user_old_phone_number_is_equal_to_new_number(self):
        data = {
            'new_phone': "+201005263977",
            "password": "secret",
        }
        response = self.client.put(self.url, data=data)
        self.assertEquals(response.status_code, 422)


class ChangePasswordAPIViewStructureAPIView(TestCase):
    def setUp(self):
        self.view = ChangePasswordAPIView

    def test_it_has_serializer_class(self):
        self.assertTrue(hasattr(self.view, 'serializer_class'))

    def test_it_has_model(self):
        self.assertTrue(hasattr(self.view, 'model'))

    def test_it_has_permission_class(self):
        self.assertTrue(hasattr(self.view, 'permission_classes'))

    def test_it_has_update_method(self):
        self.assertIn('update', self.view.__dict__)


class ChangePasswordAPIViewTestCase(TestCase):
    def setUp(self):
        self.user = UserFactory()
        self.client = APIClient()
        self.refresh = RefreshToken.for_user(self.user)
        self.client.credentials(HTTP_AUTHORIZATION='Bearer {}'.format(self.refresh.access_token))
        self.url = reverse('change_password_api')
        self.data= {
            "new_password1": '12345678Aa',
            "new_password2": '12345678Aa',
            "old_password": 'secret'
        }
        self.password = self.user.password

    def test_it_uses_change_password_serializer(self):
        self.assertIs(ChangePasswordAPIView.serializer_class, ChangePasswordSerializer)

    def test_it_returns_401_when_user_is_unauthenticated(self):
        self.client.logout()
        response = self.client.put(self.url, data={})
        self.assertEquals(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_it_returns_422_invalid_data(self):
        response = self.client.put(self.url, data={})
        self.assertEqual(response.status_code, status.HTTP_422_UNPROCESSABLE_ENTITY)

    def test_it_returns_200_on_update(self):
        response = self.client.put(self.url, data=self.data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_it_updates_password(self):
        self.client.put(self.url, data=self.data)
        self.user.refresh_from_db()
        self.assertNotEqual(self.user.password, self.password)



from django.conf import settings
from django.contrib.auth.tokens import PasswordResetTokenGenerator
from django.contrib.sites.shortcuts import get_current_site
from django.core.mail import send_mail
from django.template.loader import render_to_string
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode
from django.utils.translation import gettext as _


class TokenGenerator(PasswordResetTokenGenerator):
    def _make_hash_value(self, user, timestamp):
        return (
                str(user.pk) + str(timestamp) + str(user.is_active)
        )


account_activation_token = TokenGenerator()


def send_mail_confirmation(request, user):
    current_site = get_current_site(request)
    mail_subject = _('Activate your account.')
    message = render_to_string('dj_accounts/confirm_email_template.html', {
        'user': user,
        'domain': current_site.domain,
        'uid': urlsafe_base64_encode(force_bytes(user.pk)),
        'token': account_activation_token.make_token(user),
    })
    send_mail(mail_subject, message, settings.EMAIL_HOST_USER, [user.email])



from ..serializers import LogoutSerializer, RegisterSerializer, UpdateUserDataSerializer, UpdateEmailSerializer, \
    UpdatePhoneNumberSerializer, ChangePasswordSerializer
from ..tests.factories import UserFactory
from django.contrib.auth import password_validation
from django.test import TestCase
from django.test.client import RequestFactory
from django.utils.translation import gettext as _
from rest_framework import serializers


class RegistrationSerializerTestCase(TestCase):
    def setUp(self):
        self.serializer = RegisterSerializer()

    def test_it_extends_drf_serializer(self):
        self.assertTrue(issubclass(RegisterSerializer, serializers.ModelSerializer))

    def test_it_has_email_field(self):
        self.assertIn('email', self.serializer.fields)

    def test_email_field_is_instance_of_email_field(self):
        self.assertIsInstance(self.serializer.fields['email'], serializers.CharField)

    def test_email_field_is_required(self):
        self.assertTrue(self.serializer.fields['email'].required)

    def test_email_field_max_length_is_100(self):
        self.assertEquals(self.serializer.fields['email'].max_length, 100)

    def test_email_field_help_text(self):
        self.assertEquals(self.serializer.fields['email'].help_text, _("Required, please provide your email"))

    def test_it_has_password1_field(self):
        self.assertIn('password1', self.serializer.fields)

    def test_password1_field_is_instance_of_password1_field(self):
        self.assertIsInstance(self.serializer.fields['password1'], serializers.CharField)

    def test_password1_field_is_required(self):
        self.assertTrue(self.serializer.fields['password1'].required)

    def test_password1_field_help_text(self):
        self.assertEquals(self.serializer.fields['password1'].help_text,
                          password_validation.password_validators_help_text_html())

    def test_it_has_password2_field(self):
        self.assertIn('password2', self.serializer.fields)

    def test_password2_field_is_instance_of_password2_field(self):
        self.assertIsInstance(self.serializer.fields['password2'], serializers.CharField)

    def test_password2_field_is_required(self):
        self.assertTrue(self.serializer.fields['password2'].required)

    def test_password2_field_help_text(self):
        self.assertEquals(self.serializer.fields['password2'].help_text,
                          _("Enter the same password as before, for verification."))

    def test_it_has_username_field(self):
        self.assertIn('username', self.serializer.fields)

    def test_username_field_is_instance_of_char_field(self):
        self.assertIsInstance(self.serializer.fields['username'], serializers.CharField)

    def test_username_field_is_required(self):
        self.assertTrue(self.serializer.fields['username'].required)

    def test_username_field_max_length_is_100(self):
        self.assertEquals(self.serializer.fields['username'].max_length, 100)

    def test_username_field_help_text(self):
        self.assertEquals(self.serializer.fields['username'].help_text, _("Required, please provide your username"))

    def test_it_has_phone_field(self):
        self.assertIn('phone', self.serializer.fields)

    def test_phone_field_is_instance_of_char_field(self):
        self.assertIsInstance(self.serializer.fields['phone'], serializers.CharField)

    def test_phone_field_is_required(self):
        self.assertTrue(self.serializer.fields['phone'].required)

    def test_phone_field_max_length_is_100(self):
        self.assertEquals(self.serializer.fields['phone'].max_length, 100)

    def test_phone_field_help_text(self):
        self.assertEquals(self.serializer.fields['phone'].help_text, _("Required, please provide your phone number"))


class LogoutSerializerStructureTestCase(TestCase):
    def setUp(self):
        self.serializer = LogoutSerializer(data={'refresh': 'test'})

    def test_it_has_refresh_field(self):
        self.assertIn('refresh', self.serializer.fields)

    def test_refresh_field_help_text(self):
        self.assertEquals(self.serializer.fields['refresh'].help_text, _("Required, please provide your refresh token"))

    def test_refresh_field_is_instance_of_char_field(self):
        self.assertIsInstance(self.serializer.fields['refresh'], serializers.CharField)

    def test_refresh_field_is_required(self):
        self.assertTrue(self.serializer.fields['refresh'].required)

    def test_it_has_save_method(self):
        self.assertTrue(hasattr(self.serializer, 'save'))


# profile serializer tests

class UpdateUserDataSerializerStructureTestCase(TestCase):
    def setUp(self):
        self.serializer = UpdateUserDataSerializer(data={})

    def test_it_has_meta_class(self):
        self.assertIsInstance(self.serializer.Meta, type)

    def test_it_has_model_class_in_meta(self):
        self.assertTrue(hasattr(self.serializer.Meta, 'model'))

    def test_it_has_fields_in_meta(self):
        self.assertTrue(hasattr(self.serializer.Meta, 'fields'))

    def test_it_has_first_name_field(self):
        self.assertIn('first_name', self.serializer.fields)

    def test_it_has_last_name_field(self):
        self.assertIn('last_name', self.serializer.fields)


class UpdateEmailSerializerStructureTestCase(TestCase):
    def setUp(self):
        self.serializer = UpdateEmailSerializer(data={})

    def test_it_has_meta_class(self):
        self.assertIsInstance(self.serializer.Meta, type)

    def test_it_has_model_class_in_meta(self):
        self.assertTrue(hasattr(self.serializer.Meta, 'model'))

    def test_it_has_fields_in_meta(self):
        self.assertTrue(hasattr(self.serializer.Meta, 'fields'))

    def test_it_has_email_field(self):
        self.assertIn('email', self.serializer.fields)

    def test_it_has_password_field(self):
        self.assertIn('password', self.serializer.fields)


class UpdatePhoneSerializerStructureTestCase(TestCase):
    def setUp(self):
        self.serializer = UpdatePhoneNumberSerializer(data={})

    def test_it_has_meta_class(self):
        self.assertIsInstance(self.serializer.Meta, type)

    def test_it_has_model_class_in_meta(self):
        self.assertTrue(hasattr(self.serializer.Meta, 'model'))

    def test_it_has_fields_in_meta(self):
        self.assertTrue(hasattr(self.serializer.Meta, 'fields'))

    def test_it_has_phone_field(self):
        self.assertIn('phone', self.serializer.fields)

    def test_it_has_password_field(self):
        self.assertIn('password', self.serializer.fields)


class ChangePasswordSerializerStructureTestCase(TestCase):
    def setUp(self):
        self.serializer = ChangePasswordSerializer()

    def test_it_has_model_class_in_meta(self):
        self.assertTrue(hasattr(self.serializer.Meta, 'model'))

    def test_it_has_fields_in_meta(self):
        self.assertTrue(hasattr(self.serializer.Meta, 'fields'))

    def test_it_has_new_password1_field(self):
        self.assertIn('new_password1', self.serializer.fields)

    def test_it_has_new_password2_field(self):
        self.assertIn('new_password2', self.serializer.fields)

    def test_it_has_old_password_field(self):
        self.assertIn('old_password', self.serializer.fields)

    def test_it_has_form_attribute(self):
        self.assertTrue(hasattr(self.serializer, 'form'))

    def test_it_has_method_validate(self):
        self.assertTrue(hasattr(self.serializer, 'validate'))

    def test_it_has_method_save(self):
        self.assertTrue(hasattr(self.serializer, 'save'))


class ChangePasswordSerializerTestCase(TestCase):
    def setUp(self):
        self.request = RequestFactory()
        self.user = UserFactory()
        self.request.user = self.user
        self.old_password = self.user.password
        self.data = {
            "new_password1": "12345678Aa",
            "new_password2": "12345678Aa",
            "old_password": "secret"
        }

    def test_it_change_password(self):
        serializer = ChangePasswordSerializer(data=self.data, context={'request': self.request})
        serializer.is_valid(raise_exception=True)
        serializer.save()
        self.user.refresh_from_db()
        self.assertNotEqual(self.old_password, self.user.password)



import importlib
import sys
import traceback
from pyexpat.errors import messages

from django.conf import settings
from django.contrib import messages
from django.contrib.auth import login, get_user_model
from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.auth.views import LoginView as BaseLoginView
from django.shortcuts import render, redirect
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.http import urlsafe_base64_decode
from django.utils.timezone import now
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy as _
from django.views import View

from .forms import MultipleLoginForm, VerifyPhoneForm
from .forms import UpdatePhoneNumberForm, UpdateEmailForm
from .utils import account_activation_token, send_mail_confirmation
from .verify_phone import VerifyPhone


class UpdateProfileView(LoginRequiredMixin, View):
    def get_form_class(self):
        form_class = getattr(settings, "PROFILE_FORM", 'dj_accounts.forms.UserChangeForm')
        if type(form_class) is str:
            form_class_split = form_class.split('.')
            class_name = form_class_split[-1:][0]
            module_name = form_class_split[:-1]
            return getattr(importlib.import_module('.'.join(module_name)), class_name)
        return form_class

    def get(self, request, *args, **kwargs):
        return render(request, 'dj_accounts/update_user_data_form.html', {
            'form': self.get_form_class(),
            "page_title": _("Update User Info")
        })

    def post(self, request, *args, **kwargs):
        form = self.get_form_class()(request.POST, request.FILES)
        if form.is_valid():
            form.save()
            messages.success(request, _('your profile has been updated successfully'))
            return redirect(settings.LOGIN_REDIRECT_URL)
        return render(request, 'dj_accounts/update_user_data_form.html', {
            'form': form
        })


class ChangeEmailView(LoginRequiredMixin, View):
    def get(self, request, *args, **kwargs):
        return render(request, 'dj_accounts/update_email_form.html', {
            'form': UpdateEmailForm(),
            "page_title": _("Update Email")
        })

    def post(self, request, *args, **kwargs):
        user = request.user
        form = UpdateEmailForm(user, request.POST)
        if form.is_valid():
            form.save()
            messages.success(request, 'your email has been updated successfully')
            return redirect(settings.LOGIN_REDIRECT_URL)

        return render(request, 'dj_accounts/update_email_form.html', {
            'form': form
        })


class ChangePhoneView(LoginRequiredMixin, View):
    def get(self, request, *args, **kwargs):
        return render(request, 'dj_accounts/update_phone_number_form.html', {
            'form': UpdatePhoneNumberForm(),
            "page_title": _("Update Phone Number")
        })

    def post(self, request, *args, **kwargs):
        user = request.user
        form = UpdatePhoneNumberForm(user, request.POST)
        if form.is_valid():
            request.user.phone_verified_at = None
            request.user.save()
            return redirect(settings.LOGIN_REDIRECT_URL)

        return render(request, 'dj_accounts/update_phone_number_form.html', {
            "form": form
        })


UserModel = get_user_model()


class LoginView(BaseLoginView):
    template_name = 'dj_accounts/login.html'
    redirect_authenticated_user = True

    def get_form_class(self):
        if getattr(settings, 'MULTIPLE_AUTHENTICATION_ACTIVE', False):
            return MultipleLoginForm
        return getattr(settings, 'LOGIN_FORM', AuthenticationForm)


class RegisterView(View):
    def get_form_class(self):
        form_class = getattr(settings, "REGISTER_FORM", 'dj_accounts.forms.UserCreationForm')
        if type(form_class) is str:
            form_class_split = form_class.split('.')
            class_name = form_class_split[-1:][0]
            module_name = form_class_split[:-1]
            return getattr(importlib.import_module('.'.join(module_name)), class_name)
        else:
            return form_class

    def get(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            return redirect(settings.LOGIN_REDIRECT_URL)
        return render(request, 'dj_accounts/register.html', {
            "form": self.get_form_class()()
        })

    def post(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            return redirect(settings.LOGIN_REDIRECT_URL)

        form = self.get_form_class()(request.POST)
        if form.is_valid():
            user = form.save()
            login(self.request, user)

            try:
                send_mail_confirmation(request, user)
            except Exception as e:
                parts = ["Traceback (most recent call last):\n"]
                parts.extend(traceback.format_stack(limit=25)[:-2])
                parts.extend(traceback.format_exception(*sys.exc_info())[1:])
                print("".join(parts))

            try:
                VerifyPhone().send(user.phone)
            except Exception as e:
                parts = ["Traceback (most recent call last):\n"]
                parts.extend(traceback.format_stack(limit=25)[:-2])
                parts.extend(traceback.format_exception(*sys.exc_info())[1:])
                print("".join(parts))

            if 'next' in request.POST:
                return redirect(request.POST.get('next'))
            return redirect(settings.LOGIN_REDIRECT_URL)

        return render(self.request, 'dj_accounts/register.html', {
            "form": form
        })


# phone verification
class VerifyPhoneView(LoginRequiredMixin, View):
    def get(self, request, *args, **kwargs):
        if request.user.phone_verified_at is not None:
            return redirect(settings.LOGIN_REDIRECT_URL)

        return render(request, 'dj_accounts/verify_phone.html', {
            "form": VerifyPhoneForm(user=request.user)
        })

    def post(self, request, *args, **kwargs):
        if request.user.phone_verified_at is not None:
            return redirect(settings.LOGIN_REDIRECT_URL)

        form = VerifyPhoneForm(request.POST, user=request.user)
        if form.is_valid():
            request.user.phone_verified_at = now()
            request.user.save()
            return redirect(reverse("phone-verification-complete"))

        return render(request, 'dj_accounts/verify_phone.html', {"form": form})


class PhoneVerificationCompleteView(LoginRequiredMixin, View):
    def get(self, request):
        return render(request, "dj_accounts/phone_verification_complete.html")


class ResendPhoneConfirmationView(LoginRequiredMixin, View):

    def get(self, request, *args, **kwargs):
        try:
            VerifyPhone().send(request.user.phone)
        except Exception as e:
            parts = ["Traceback (most recent call last):\n"]
            parts.extend(traceback.format_stack(limit=25)[:-2])
            parts.extend(traceback.format_exception(*sys.exc_info())[1:])
            print("".join(parts))

        messages.success(request, _("A new confirmation code has been sent to your phone"))
        return redirect(reverse("verify-phone"))


# email verification

class VerifyEmailView(LoginRequiredMixin, View):
    def get(self, request, uidb64, token):
        try:
            uid = force_text(urlsafe_base64_decode(uidb64))
            user = UserModel.objects.get(pk=uid)
        except(TypeError, ValueError, OverflowError, UserModel.DoesNotExist):
            user = None
        if user is not None and account_activation_token.check_token(user, token):
            user.email_verified_at = now()
            user.save()

        return redirect('email-verification-complete')


class EmailVerificationCompleteView(LoginRequiredMixin, View):
    def get(self, request):
        return render(request, 'dj_accounts/email_verification_complete.html', {
            'verified': request.user.email_verified_at
        })


class ResendEmailConfirmationLinkView(View):
    def get(self, request, *args, **kwargs):
        try:
            send_mail_confirmation(request, request.user)
        except Exception as e:
            parts = ["Traceback (most recent call last):\n"]
            parts.extend(traceback.format_stack(limit=25)[:-2])
            parts.extend(traceback.format_exception(*sys.exc_info())[1:])
            print("".join(parts))

        messages.success(request, 'email verification is sent successfully')
        return redirect(kwargs['next'])



from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend

UserModel = get_user_model()


class MultipleAuthenticationBackend(ModelBackend):
    def authenticate(self, request, *args, **kwargs):
        credentials = {}
        password = kwargs.pop('password', None)

        if not password:
            return

        if len(kwargs.keys()) == 0:
            return

        try:
            user = UserModel.objects.get(**kwargs)
        except UserModel.DoesNotExist:
            # Run the default password hasher once to reduce the timing
            # difference between an existing and a nonexistent user (#20760).
            UserModel().set_password(password)
        else:
            if user.check_password(password) and self.user_can_authenticate(user):
                return user



import importlib
import sys
import traceback

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.tokens import default_token_generator
from django.utils.encoding import force_text
from django.utils.http import urlsafe_base64_decode
from django.utils.timezone import now
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy as _
from rest_framework import status
from rest_framework.generics import UpdateAPIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework_simplejwt.tokens import RefreshToken

from .forms import UpdateEmailForm, UpdatePhoneNumberForm
from .forms import VerifyPhoneForm
from .serializers import LogoutSerializer, PasswordResetSerializer, RegisterSerializer, \
    LoginSerializer, ChangePasswordSerializer
from .utils import account_activation_token, send_mail_confirmation
from .verify_phone import VerifyPhone

UserModel = get_user_model()


class UserSignupAPIView(APIView):
    access_token = None
    refresh_token = None
    token = None
    user = None
    authentication_classes = []
    permission_classes = []

    def post(self, request, *args, **kwargs):
        serializer = RegisterSerializer(data=request.data)
        if serializer.is_valid():
            self.user = serializer.save()

            try:
                send_mail_confirmation(request, self.user)
            except Exception as e:
                parts = ["Traceback (most recent call last):\n"]
                parts.extend(traceback.format_stack(limit=25)[:-2])
                parts.extend(traceback.format_exception(*sys.exc_info())[1:])
                print("".join(parts))

            try:
                VerifyPhone().send(self.user.phone)
            except Exception as e:
                parts = ["Traceback (most recent call last):\n"]
                parts.extend(traceback.format_stack(limit=25)[:-2])
                parts.extend(traceback.format_exception(*sys.exc_info())[1:])
                print("".join(parts))

            refresh = RefreshToken.for_user(self.user)
            return Response({
                "access_token": str(refresh.access_token),
                "refresh_token": str(refresh)
            }, status=status.HTTP_201_CREATED)

        return Response(serializer.errors, status=status.HTTP_422_UNPROCESSABLE_ENTITY)


class LoginAPIView(APIView):
    authentication_classes = []
    permission_classes = []

    def post(self, request, *args, **kwargs):
        serializer = LoginSerializer(data=request.data)
        if serializer.is_valid():
            user = serializer.cache_user
            tokens = RefreshToken.for_user(user)
            return Response({
                "access_token": str(tokens.access_token),
                "refresh_token": str(tokens)
            }, status=status.HTTP_200_OK)
        return Response(serializer.errors, status=status.HTTP_422_UNPROCESSABLE_ENTITY)


class UserLogoutAPIView(APIView):
    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        serializer = LogoutSerializer(data=request.data)
        if serializer.is_valid():
            serializer.save()
            return Response(status=status.HTTP_204_NO_CONTENT)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)


class ChangePasswordAPIView(UpdateAPIView):
    serializer_class = ChangePasswordSerializer
    model = UserModel
    permission_classes = [IsAuthenticated]

    def update(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data, context={'request': request})

        if serializer.is_valid():
            serializer.save()
            return Response(status=status.HTTP_200_OK, data={'msg': _("Password updated successfully")})
        return Response(status=status.HTTP_422_UNPROCESSABLE_ENTITY, data=serializer.errors)


class PasswordResetAPIView(APIView):
    permission_classes = []
    authentication_classes = []
    email_template_name = 'dj_accounts/password_reset_email.html'
    extra_email_context = None
    from_email = None
    html_email_template_name = None
    subject_template_name = 'dj_accounts/password_reset_subject.txt'
    token_generator = default_token_generator

    def post(self, request, *args, **kwargs):
        serializer = PasswordResetSerializer(data=request.data)
        if serializer.is_valid():
            opts = {
                'use_https': self.request.is_secure(),
                'token_generator': self.token_generator,
                'from_email': self.from_email,
                'email_template_name': self.email_template_name,
                'subject_template_name': self.subject_template_name,
                'request': self.request,
                'html_email_template_name': self.html_email_template_name,
                'extra_email_context': self.extra_email_context,
            }
            serializer.save(opts=opts)
            return Response({"message": "{} {}".format(
                _('We’ve emailed you instructions for setting your password, '
                  'if an account exists with the email you entered. You should receive them shortly.'),
                _('If you don’t receive an email, please make sure you’ve entered '
                  'the address you registered with, and check your spam folder.'))},
                status=status.HTTP_200_OK)
        return Response(serializer.errors, status=status.HTTP_422_UNPROCESSABLE_ENTITY)


class VerifyPhoneAPIView(APIView):
    permission_classes = [IsAuthenticated, ]

    def post(self, request, *args, **kwargs):
        if request.user.phone_verified_at is not None:
            return Response({'message': 'this account was activated before'}, status=status.HTTP_400_BAD_REQUEST)

        form = VerifyPhoneForm(request.POST, user=request.user)
        if form.is_valid():
            request.user.phone_verified_at = now()
            request.user.save()
            return Response(status=status.HTTP_200_OK)
        return Response(form.errors, status=status.HTTP_422_UNPROCESSABLE_ENTITY)


class ResendPhoneConfirmationAPIView(APIView):
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        try:
            VerifyPhone().send(request.user.phone)
        except Exception as e:
            parts = ["Traceback (most recent call last):\n"]
            parts.extend(traceback.format_stack(limit=25)[:-2])
            parts.extend(traceback.format_exception(*sys.exc_info())[1:])
            print("".join(parts))

        return Response({"message": _('Code was resent successfully.')}, status=status.HTTP_200_OK)


# email verification
class ResendEmailConfirmationLinkAPIView(APIView):
    def get(self, request, *args, **kwargs):
        try:
            send_mail_confirmation(request, request.user)
        except Exception as e:
            parts = ["Traceback (most recent call last):\n"]
            parts.extend(traceback.format_stack(limit=25)[:-2])
            parts.extend(traceback.format_exception(*sys.exc_info())[1:])
            print("".join(parts))

        return Response({"message": _('Email activation link resent successfully')})


class VerifyEmailAPIView(APIView):
    permission_classes = (IsAuthenticated,)

    def get(self, request, uidb64, token):
        try:
            uid = force_text(urlsafe_base64_decode(uidb64))
            user = UserModel.objects.get(pk=uid)
        except(TypeError, ValueError, OverflowError, UserModel.DoesNotExist):
            user = None
        if user is not None and account_activation_token.check_token(user, token):
            user.email_verified_at = now()
            user.save()
        return Response({"message": _('Email was verified successfully.')}, status=status.HTTP_200_OK)


class UpdateProfileAPIView(APIView):
    """
    An endpoint for changing User Profile Data.
    """

    def get_serializer_class(self):
        serializer_class = getattr(settings, "PROFILE_SERIALIZER", 'dj_accounts.serializer.UpdateUserDataSerializer')
        if type(serializer_class) is str:
            form_class_split = serializer_class.split('.')
            class_name = form_class_split[-1:][0]
            module_name = form_class_split[:-1]
            return getattr(importlib.import_module('.'.join(module_name)), class_name)
        return serializer_class

    permission_classes = (IsAuthenticated,)

    def put(self, request, *args, **kwargs):
        serializer = self.get_serializer_class()(data=request.data, instance=request.user)
        if serializer.is_valid():
            serializer.save()
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        return Response(serializer.errors, status=status.HTTP_422_UNPROCESSABLE_ENTITY)


class ChangeEmailAPIView(APIView):
    """
    An endpoint for changing Email.
    """
    permission_classes = (IsAuthenticated,)

    def get_serializer_class(self):
        # todo: change this serializer to basic Update Email Form
        return getattr(settings, "UPDATE_EMAIL_FORM", UpdateEmailForm)

    def put(self, request, *args, **kwargs):
        validation_from = self.get_serializer_class()(data=request.data, user=request.user)
        if validation_from.is_valid():
            validation_from.save()
            return Response({'message': _('Email Changed Successfully')}, status=status.HTTP_201_CREATED)
        return Response(validation_from.errors, status=status.HTTP_422_UNPROCESSABLE_ENTITY)


class ChangePhoneAPIView(APIView):
    """
    An endpoint for changing Phone.
    """
    permission_classes = (IsAuthenticated,)

    def get_serializer_class(self):
        return getattr(settings, "UPDATE_PHONE_FORM", UpdatePhoneNumberForm)

    def put(self, request, *args, **kwargs):
        validated_fields = self.get_serializer_class()(data=request.data, user=request.user)
        if validated_fields.is_valid():
            validated_fields.save()
            return Response(status=status.HTTP_201_CREATED)
        return Response(validated_fields.errors, status=status.HTTP_422_UNPROCESSABLE_ENTITY)



from ...views_api import ChangeEmailAPIView, UpdateProfileAPIView, ChangePhoneAPIView
from django.urls import path

urlpatterns = [
    path('update-profile/', UpdateProfileAPIView.as_view(), name='update-profile-api'),
    path('change-email/', ChangeEmailAPIView.as_view(), name='change-email-api'),
    path('change-phone/', ChangePhoneAPIView.as_view(), name='change-phone-api'),
]
